<?php
class PaywerkPayments extends WC_Payment_Gateway {

  function configAsInt($key) {
    if (!isset($this->settings[$key])) return false;
    $value = $this->settings[$key];
    return is_numeric($value) ? (int)$value : false;
  }

  // Docs: https://docs.woocommerce.com/document/payment-gateway-api/#section-3
  function __construct() {
    // The global ID for this Payment method
    $this->id = PaywerkSettings::TEXT_DOMAIN;
    $this->supports = array('products', 'refunds');

    // Defines language used for Paywerk process
    $this->lang = preg_split('/_/', get_locale())[0];

    // Defines supported languages for Paywerk plugin
    $this->supported_langs = array('en', 'et', 'pl');

    // The Title shown on the top of the Payment Gateways Page next to all the other Payment Gateways
    $this->method_title = $this->paywerk_label("plugin.title", 'Paywerk Payments');
  
    // The description for this Payment Gateway, shown on the actual Payment options page on the backend
    $this->method_description = $this->paywerk_label("plugin.description", 'Paywerk Payments Plug-in for WooCommerce');

    $this->init_settings();
    $this->isSandbox = isset($this->settings['sandbox']) && $this->settings['sandbox'] == 'yes';
    $this->url = $this->settings['url'] ?? PaywerkSettings::DEFAULT_URL;
    $this->shopperurl = $this->settings['shopperurl'] ?? PaywerkSettings::DEFAULT_SHOPPER_PORTAL_URL;
    $this->apiKey = $this->settings['apiKey'] ?? '';
    $this->icon = $this->url . '/img/paywerk-payments-logo.svg';

    $this->max_amount = $this->configAsInt('maxAmount');
    $this->min_amount = $this->configAsInt('minAmount');

    // The title to be used for the vertical tabs that can be ordered top to bottom

    $checkoutBlockTitleValue = isset($this->settings['checkoutBlockTitle']) ? $this->settings['checkoutBlockTitle'] : 'plugin.checkout.paymentGatewayTitle1';
    if ($checkoutBlockTitleValue == 'plugin.checkout.paymentGatewayTitle4') {
      $this->title = isset($this->settings["checkoutCustomLabel{$this->lang}"]) ? $this->settings["checkoutCustomLabel{$this->lang}"] : '';
    } else {
      $this->title = $this->paywerk_label($checkoutBlockTitleValue, 'Pay now or later with card');
    }

    $this->has_fields = true;

    if ($this->enabled == 'yes') {
      add_filter('woocommerce_thankyou_order_received_text', array($this, 'order_received_text'), 10, 2);
      add_filter('woocommerce_available_payment_gateways', array($this, 'display_gateways'), 10, 1);
      add_filter('woocommerce_checkout_fields', array($this, 'trigger_update_checkout_on_change'), 10, 1);
      add_filter('woocommerce_order_item_get_formatted_meta_data', array($this, 'hide_paywerk_meta'), 10, 1);
      add_action('woocommerce_order_status_changed', array($this, 'order_status_changed'), 10, 3);
      add_action('woocommerce_add_order_item_meta', array($this, 'add_order_item_meta') , 1, 3);
    }

    if (is_admin()) {
      $this->init_form_fields();

      //do not delete this, process_admin_options is in Woo code
      add_action('woocommerce_update_options_payment_gateways_' . $this->id, array($this, 'process_admin_options'));
      add_filter('woocommerce_settings_api_sanitized_fields_' . $this->id, array($this, 'settings_updated'));
    }

    add_action('woocommerce_api_paywerk_payment_status', array($this, 'paywerk_payment_status'));
  }

  public function paywerk_label($label_key, $fallback_label) {
    $translated_label = __($label_key, $this->id);
    return in_array($this->lang, $this->supported_langs) ? ( $translated_label != $label_key ? $translated_label : $fallback_label) : $fallback_label;
  }

  public function paywerk_payment_status() {
    $inputJSON = file_get_contents('php://input');
    $parsePostBody = json_decode($inputJSON, TRUE); //convert JSON into array

    $isPost = $_SERVER['REQUEST_METHOD'] == 'POST';

    $transactionId = $isPost ? sanitize_text_field($parsePostBody['paywerkId']) : $_GET['paywerkId'];
    $orderId = $isPost ? sanitize_text_field($parsePostBody['orderReference']) : $_GET['orderReference'];
    $order = wc_get_order($orderId);

    $this->processPaywerkStatusChange($order, $transactionId);
  }

  public function display_gateways($available_gateways) {
    return array_filter($available_gateways, function($g) {
      return !($g instanceof PaywerkPayments) || !$g->needs_setup();
    });
  }

  public function needs_setup() {
    return !$this->apiKey;
  }

  public function get_icon() {
    switch (strtolower($this->lang)) {
      case 'et':
        $url = 'https://paywerk.co/et/kuidas-see-tootab/';
        break;
      case 'pl':
        $url = 'https://paywerk.co/pl/jak-to-dziala/';
        break;
      default:
        $url = 'https://paywerk.co/how-it-works/';
        break;
    }

    $html = ' <div style="width: 5px;"></div><div class="pw-image"><a target="_blank" href="' . $url . '">' . $this->paywerk_label('plugin.checkout.howItWorks', 'How it works?') . '</a><div class="flex-grow-1"></div><img src="' . plugin_dir_url(__FILE__) . '/img/paywerk-payments-logo.svg" alt="Paywerk Payments"></div>';
    return apply_filters('woocommerce_gateway_icon', $html, $this->id);
  }

  // Docs: https://docs.woocommerce.com/document/settings-api/
  public function init_form_fields() {
    $predefinedStatuses = wc_get_order_statuses();
    $statusesWithDoNotSend = $predefinedStatuses;
    $statusesWithDoNotSend[''] = $this->paywerk_label('plugin.admin.statusActions.notInUse', 'Do not send');

    $this->form_fields = [
      'apiKey' => [
        'title'		=> $this->paywerk_label('plugin.admin.apikey.label', 'Paywerk API Key'),
        'type'		=> 'text',
        'desc_tip' => $this->paywerk_label('plugin.admin.apikey.description', 'You can generate API keys in Merchant Portal'),
      ],
      'sandbox' => [
        'title'		=> $this->paywerk_label('plugin.admin.sandbox.label', 'Sandbox'),
        'label'		=> $this->paywerk_label('plugin.admin.sandbox.description', 'Use Paywerk\'s sandbox environment to test functionality without real money transfers'),
        'type'		=> 'checkbox',
        'default'	=> 'no',
      ],
      'url' => [
        'type'		=> 'hidden',
        'default'	=> PaywerkSettings::DEFAULT_URL
      ],
      'minAmount' => [
        'title'		=> $this->paywerk_label('plugin.admin.minAmount.label', 'Minimum order amount'),
        'type'		=> 'text',
        'desc_tip' => $this->paywerk_label('plugin.admin.minAmount.description', 'Minimum order amount supported by payment method'),
      ],
      'maxAmount' => [
        'title'		=> $this->paywerk_label('plugin.admin.maxAmount.label', 'Maximum order amount'),
        'type'		=> 'text',
        'desc_tip' => $this->paywerk_label('plugin.admin.maxAmount.description', 'Maximum order amount supported by payment method'),
      ],
      'completedStatus' => [
        'title'		=> $this->paywerk_label('plugin.admin.paymentCompleted.label', 'Payment completed status'),
        'type'		=> 'select',
        'options'     => $predefinedStatuses,
        'desc_tip' => $this->paywerk_label('plugin.admin.paymentCompleted.description', 'After successful Paywerk payment set order status to this'),
        'default'	=> 'wc-processing',
      ],
      'fulfilledStatus' => [
        'title'		=> $this->paywerk_label('plugin.admin.fulfilledStatus.label', 'Order fulfilled status'),
        'type'		=> 'select',
        'options'     => $statusesWithDoNotSend,
        'desc_tip' => $this->paywerk_label('plugin.admin.fulfilledStatus.description', 'After this status fulfil whole order request sent to Paywerk'),
        'default'	=> 'wc-completed',
      ],
      'refundedStatus' => [
        'title'		=> $this->paywerk_label('plugin.admin.refundedStatus.label', 'Order refunded status'),
        'type'		=> 'select',
        'options'     => $statusesWithDoNotSend,
        'desc_tip' => $this->paywerk_label('plugin.admin.refundedStatus.description', 'After this status refund whole order request sent to Paywerk'),
        'default'	=> 'wc-refunded',
      ],
      'checkoutBlockTitle' => [
        'title'		=> $this->paywerk_label('plugin.admin.checkout.block.label', 'Checkout block title'),
        'type'		=> 'select',
        'options'     => array(
          'plugin.checkout.paymentGatewayTitle1' => $this->paywerk_label('plugin.checkout.paymentGatewayTitle1', 'Pay now or later with card'),
          'plugin.checkout.paymentGatewayTitle2' => $this->paywerk_label('plugin.checkout.paymentGatewayTitle2', 'Pay later with card'),
          'plugin.checkout.paymentGatewayTitle3' => $this->paywerk_label('plugin.checkout.paymentGatewayTitle3', 'Pay by card'),
          'plugin.checkout.paymentGatewayTitle4' => $this->paywerk_label('plugin.checkout.paymentGatewayTitle4', 'Custom title'),
        ),
        'desc_tip' => $this->paywerk_label('plugin.admin.checkout.block.description', 'Public title for payment block'),
        'default'	=> 'plugin.checkout.paymentGatewayTitle2',
      ],
    ];

    if (isset($this->settings['checkoutBlockTitle']) && $this->settings['checkoutBlockTitle'] == 'plugin.checkout.paymentGatewayTitle4') {
      foreach ($this->supported_langs as $language) {
        $this->form_fields["checkoutCustomLabel{$language}"] =  [
          'title'		=> $this->paywerk_label("plugin.admin.checkout.block.label.{$language}", 'Checkout block title for '.strtoupper($language)),
          'type'		=> 'text',
          'desc_tip' => $this->paywerk_label('plugin.admin.checkout.block.description', 'Public title for payment block'),
        ];
      };
    }

    $this->form_fields["promotionsTitle"] =  [
      'title'		=> $this->paywerk_label('plugin.admin.promotionsTitle.label', 'Promotions'),
      'type'		=> 'hidden',
    ];

    $this->form_fields["promotionsVisibility"] =  [
      'title'		=> $this->paywerk_label('plugin.admin.promotionsVisibility.label', 'Promotion visibility'),
      'type'		=> 'select',
      'options'     => array(
        'notVisible' => $this->paywerk_label('plugin.promotions.location.none', 'Not shown'),
        'woocommerce_after_add_to_cart_button' => $this->paywerk_label('plugin.promotions.location.afterAddToCartButton', 'After add to cart button'),
        'woocommerce_single_product_summary' => $this->paywerk_label('plugin.promotions.location.singleProductSummary', 'Single product summary'),
        'woocommerce_after_single_product_summary' => $this->paywerk_label('plugin.promotions.location.afterSingleProductSummary', 'After single product summary'),
      ),
      'desc_tip' => $this->paywerk_label('plugin.admin.promotionsVisibility.description', 'Where to show promotions panel'),
      'default'	=> 'notVisible',
    ];

    $this->form_fields["promotionsMessage"] =  [
      'title'		=> $this->paywerk_label('plugin.admin.promotionsMessage.label', 'Promotion message'),
      'type'		=> 'select',
      'options'     => array(
        "INVOICE_30_1" => $this->paywerk_label('plugin.promotions.location.INVOICE_30_1', 'Pay in 30 days - variant 1'),
        "INVOICE_30_2" => $this->paywerk_label('plugin.promotions.location.INVOICE_30_2', 'Pay in 30 days - variant 1'),
        "SLICE_3_1" =>    $this->paywerk_label('plugin.promotions.location.SLICE_3_1', 'Pay in 3 parts - variant 1'),
        "SLICE_3_2" =>    $this->paywerk_label('plugin.promotions.location.SLICE_3_2', 'Pay in 3 parts - variant 2'),
        "SLICE_3_3" =>    $this->paywerk_label('plugin.promotions.location.SLICE_3_3', 'Pay in 3 parts - variant 3'),
        "SLICE_6_1" =>    $this->paywerk_label('plugin.promotions.location.SLICE_6_1', 'Pay in 6 parts - variant 1'),
        "SLICE_6_2" =>    $this->paywerk_label('plugin.promotions.location.SLICE_6_2', 'Pay in 6 parts - variant 2'),
        "SLICE_6_3" =>    $this->paywerk_label('plugin.promotions.location.SLICE_6_3', 'Pay in 6 parts - variant 3'),
        "SLICE_12_1" =>   $this->paywerk_label('plugin.promotions.location.SLICE_12_1', 'Pay in 12 parts - variant 1'),
        "SLICE_12_2" =>   $this->paywerk_label('plugin.promotions.location.SLICE_12_2', 'Pay in 12 parts - variant 2'),
        "SLICE_12_3" =>   $this->paywerk_label('plugin.promotions.location.SLICE_12_3', 'Pay in 12 parts - variant 3'),
      ),
      'desc_tip' => $this->paywerk_label('plugin.admin.promotionsMessage.description', 'Which text is shown on promotion panel. Note that not all products may be accessible to your company but message is still shown based on this selection.'),
      'default'	=> 'SLICE_3_3',
    ];

    $this->form_fields["promotionsTheme"] =  [
      'title'		=> $this->paywerk_label('plugin.admin.promotionsTheme.label', 'Theme'),
      'type'		=> 'select',
      'options'     => array(
        'light' => $this->paywerk_label('plugin.promotions.theme.light', 'Light'),
        'dark' => $this->paywerk_label('plugin.promotions.theme.dark', 'Dark'),
      ),
      'desc_tip' => $this->paywerk_label('plugin.admin.promotionsTheme.description', 'Colour palette used in the area where promotion is shown'),
      'default'	=> 'light',
    ];
  }

  function generate_settings_html($form_fields = [], $echo = true) {
    return (isset($this->settings['error']) ? '<p style="color: red">' . $this->settings['error'] . '</p>' : '') .
      parent::generate_settings_html($form_fields, $echo);
  }

  public function settings_updated($settings) {
    $settings = PaywerkSettings::correctPaywerkUrl($settings);
    return $settings;
  }

  public function payment_fields() {
    if (!WC()->session->__isset("customerCheckoutFields")) {
      return;
    }

    if (!$this->isMinAmountInRange()) {
      return;
    }

    $response = $this->makeOrderRequest();
    if (!empty($response->errors)) return;

    // order request cannot be updated anymore
    if (!empty($response) && $response['response']['code'] == 409) {
      WC()->session->set("orderReference", uniqid('6'));
      $response = $this->makeOrderRequest();
    }

    if ($this->isSandbox) {
      echo '<!--';
      print_r($response);
      echo '-->';
    }

    $offer = (isset($response['response']) && $response['response']['code'] == 200) ? json_decode($response['body']) : null;
    $paywerkId = $offer ? $offer->paywerkId : '';

    if (!$paywerkId) echo $this->paywerk_label('plugin.checkout.productsNotAvailable', 'Paywerk products are not available')
    ?>

    <paywerk-products paywerkId="<?php echo esc_attr($paywerkId) ?>"/>

    <input type="hidden" name="paywerkUrl">
    <script type="module" async>
      <?php if ($paywerkId) { ?>

      function registerCheckoutRefreshedHandler() {
        jQuery(document).one('updated_checkout', function(e) {
          window.paymentsLoaded = new Date().getTime()
          checkPaywerkProducts(3000)
        })
      }

      registerCheckoutRefreshedHandler()

      let paywerkProducts = document.querySelector('paywerk-products');
      paywerkProducts.addEventListener('product-selected', function (e) {
        e.preventDefault()
        document.querySelector('[name=paywerkUrl]').value = e.detail.productUrl
      })

      paywerkProducts.addEventListener('products-loaded', function (e) {
        e.preventDefault()
        window.productsLoaded = new Date().getTime()
      })

      function checkPaywerkProducts(time) {
        setTimeout(function () {
          if (window.productsLoaded < window.paymentsLoaded ) {
            console.warn("Force reloading Paywerk products")
            paywerkProducts.reloadProducts()
            registerCheckoutRefreshedHandler()
          }
        }, time)
      }

      jQuery('li.payment_method_paywerk-payments').one('click', function() {
        window.paymentsLoaded = new Date().getTime()
        checkPaywerkProducts(500)
      })

      jQuery('input[name^=billing]').on('change', function(e) {
        e.stopImmediatePropagation();
        jQuery('body').trigger('update_checkout');
      })

      <?php } ?>
      </script>
    <?php
  }

  private function to_iso(string $date) {
    return str_replace(' ', 'T', $date);
  }

  public function validate_fields() {
    if (!sanitize_text_field($_POST["paywerkUrl"])) {
      wc_add_notice($this->paywerk_label('plugin.checkout.chooseProduct', 'Please choose Paywerk payment option'), 'error');
      return false;
    }
  }

  // Submit payment and handle response
  public function process_payment($order_id) {
    $order = wc_get_order($order_id);
    $orderReference = $order->get_order_number();
    $successUrl = urlencode($this->get_return_url($order));
    $failedUrl = urlencode($order->get_cancel_order_url_raw());
    WC()->session->set("orderReference", uniqid('4'));
    return array(
      'result'   => 'success',
      'redirect' => sanitize_url($_POST["paywerkUrl"]) . "?orderReference=$orderReference&successUrl=$successUrl&failedUrl=$failedUrl"
    );
  }

  public function order_received_text($text, WC_Order $order) {
    if ($order && $this->id === $order->get_payment_method()) {
      $this->processPaywerkStatusChange($order, sanitize_text_field($_GET['paywerkId']));
      return $text . '<br>'.sprintf($this->paywerk_label('plugin.checkout.success.goToPaywerkLink', '%1$sLog into your Paywerk Shopper Portal%2$s to view transaction details and manage subsequent payments.'), '<a target="_blank" href="' . esc_url($this->shopperurl) . '/?lang=' . esc_attr($this->lang) . '">', '</a>');
    }
  }

  public function order_status_changed($order_id, $old_status, $new_status) {
    $order = wc_get_order($order_id);

    $fulfilledStatus = isset($this->settings['fulfilledStatus']) && !empty($this->settings['fulfilledStatus']) ? str_replace('wc-', '', $this->settings['fulfilledStatus']) : 'wc-completed';
    $refundedStatus = isset($this->settings['refundedStatus']) && !empty($this->settings['refundedStatus']) ? str_replace('wc-', '', $this->settings['refundedStatus']) : 'wc-refunded';

    if ($new_status == $fulfilledStatus) $this->fulfilAll($order);
    if ($new_status == $refundedStatus) $this->refundAll($order);
  }

  public function add_order_item_meta($item_id, $cart_item, $cart_item_key) {
    wc_add_order_item_meta($item_id, '_item_cart_key' , $cart_item_key, true);
  }

  public function hide_paywerk_meta($formatted_meta) {
    $new_metas = [];
    foreach($formatted_meta as $key => $meta) {
      if (isset( $meta->key ) && ! in_array($meta->key, ['_item_cart_key'])) {
        $new_metas[$key] = $meta;
      }
    }
    return $new_metas;
  }

  public function fulfilAll($order) {
    if ($order && $this->id === $order->get_payment_method()) {
      $url = $this->url . '/api/v1/orders/' . $order->get_transaction_id() . '/fulfil-all';
      $response = wp_remote_post($url, [
        'timeout' => 10,
        'headers' => [
          'Content-Type' => 'application/json',
          'Accept' => 'application/json',
          'Accept-Language' => $this->lang,
          'Authorization' => 'Bearer ' . $this->apiKey
        ]
      ]);

      if ($response['response']['code'] >= 300) {
        $paywerkError = $this->getPaywerkError($response);
        WC_Admin_Notices::add_custom_notice('pw_fulfil_notice', $this->paywerk_label('plugin.admin.fulfil.failed', 'Failed to fulfil order in Paywerk') . ': ' . $paywerkError);
      }

      if ($response['response']['code'] < 300) {
        WC_Admin_Notices::add_custom_notice('pw_fulfil_success', $this->paywerk_label('plugin.admin.fulfil.success', 'Order marked as fulfilled in Paywerk'));
      }
    }
  }

  public function refundAll($order) {
    if ($order && $this->id === $order->get_payment_method()) {
      $url = $this->url . '/api/v1/orders/' . $order->get_transaction_id() . '/refund-all';
      $response = wp_remote_post($url, [
        'timeout' => 10,
        'headers' => [
          'Content-Type' => 'application/json',
          'Accept' => 'application/json',
          'Accept-Language' => $this->lang,
          'Authorization' => 'Bearer ' . $this->apiKey
        ]
      ]);

      if ($response['response']['code'] >= 300) {
        $paywerkError = $this->getPaywerkError($response);
        WC_Admin_Notices::add_custom_notice('pw_refund_notice', $this->paywerk_label('plugin.admin.refund.failed', 'Failed to refund order in Paywerk') . ': ' . $paywerkError);
      }

      if ($response['response']['code'] < 300) {
        WC_Admin_Notices::add_custom_notice('pw_refund_success', $this->paywerk_label('plugin.admin.refund.success', 'Order fully refunded in Paywerk'));
      }
    }
  }

  public function trigger_update_checkout_on_change($fields) {
    $fields['billing']['billing_first_name']['class'][] = 'update_totals_on_change';
    $fields['billing']['billing_last_name']['class'][] = 'update_totals_on_change';
    $fields['billing']['billing_phone']['class'][] = 'update_totals_on_change';
    $fields['billing']['billing_email']['class'][] = 'update_totals_on_change';
    return $fields;
  }

  private function processPaywerkStatusChange(WC_Order $order, $transactionId) {
    if ($order->get_transaction_id()) return;

    $checkUrl = $this->url . '/api/v1/orders/' . $transactionId . '/status';

    $headers = [
      'Content-Type' => 'application/json',
      'Accept' => 'application/json',
      'Accept-Language' => $this->lang,
      'Authorization' => 'Bearer ' . $this->apiKey,
    ];
    $response = wp_remote_get($checkUrl, array('timeout' => 10, 'headers' => $headers));
    if ($response['response']['code'] >= 300) {
      $paywerkError = $this->getPaywerkError($response);
      throw new Exception($this->paywerk_label('plugin.error.technical', 'We are currently experiencing problems trying to connect to Paywerk. Please try again.') . ': ' . $paywerkError, $this->id);
    }

    $res = json_decode(wp_remote_retrieve_body($response));
    if ($res->status == 'CONFIRMED' && abs($res->amount - $order->get_total()) / $res->amount < 0.001) {
      $order->add_order_note($this->paywerk_label('plugin.checkout.success.payment', 'Paywerk payment completed.'));
      $order->payment_complete($transactionId);

      if (isset($this->settings['completedStatus']) && $this->settings['completedStatus'] != 'wc-processing') {
        $orderCompletedStatus = !empty($this->settings['completedStatus']) ? $this->settings['completedStatus'] : 'wc-processing';
        $order->update_status($orderCompletedStatus);
      }
    }
  }


  public function process_refund($order_id, $amount = null, $reason = '') {
    $order = wc_get_order($order_id);

    if ($order && $this->id === $order->get_payment_method()) {
      if (empty($order->get_transaction_id())) {
        wp_send_json_error(array('error' => $this->paywerk_label('plugin.error.orderNotFound', 'Order not found')));
        return;
      }

      $refunds = $order->get_refunds();
      $refund = $refunds[0];

      $items = array_map(function (WC_Order_Item_Product $i) {
        $item = $i->get_data();

        $refundedItemId = $i->get_meta('_refunded_item_id');
        $itemCartKey = isset($refundedItemId) && !empty($refundedItemId) ? wc_get_order_item_meta($refundedItemId, '_item_cart_key') : null;

        $variationID = $item['variation_id'];
        $variationID = $variationID != 0 ? $variationID : $item['product_id'];
        $refundQty = abs($item['quantity']);
        $unitPrice = abs($item['total'] + $item['total_tax']) / max($refundQty, 1);
        return [
          'quantity' => $refundQty,
          'unitPrice' => $unitPrice,
          'sku' => isset($itemCartKey) && !empty($itemCartKey) ? $variationID . '_' . $itemCartKey : $variationID,
          'refundType' => 'AUTO'
        ];
      }, array_values($refund->items));

      $shippingToRefund = abs($refund->shipping_total);
      if ($shippingToRefund > 0) {
        $items[] = [
          'quantity' => 1,
          'sku' => 'shipping',
          'unitPrice' => $shippingToRefund,
          'refundType' => 'AUTO'
        ];
      }

      $url = $this->url . '/api/v1/orders/' . $order->get_transaction_id() . '/refund';
      $response = wp_remote_post($url, [
        'timeout' => 10,
        'headers' => [
          'Content-Type' => 'application/json',
          'Accept' => 'application/json',
          'Accept-Language' => $this->lang,
          'Authorization' => 'Bearer ' . $this->apiKey
        ],
        'body' => json_encode($items)
      ]);

      if ($response['response']['code'] >= 300) {
        $paywerkError = $this->getPaywerkError($response);
        wp_send_json_error(array('error' => $this->paywerk_label('plugin.admin.refund.failed', 'Failed to refund order in Paywerk') . ': ' . $paywerkError));
        return false;
      }
    }

    return true;
  }

  private function getPaywerkError($response) {
    $responseBody = json_decode($response['body']);
    return $responseBody ? $responseBody->message ?? $responseBody->reason : $this->paywerk_label('plugin.error.technical', 'We are currently experiencing problems trying to connect to Paywerk. Please try again.');
  }

  /**
   * @return mixed
   */
  public function makeOrderRequest() {
    $customer = WC()->session->customer;
    $cart = WC()->cart;
    $customerCheckoutFields = WC()->session->get("customerCheckoutFields");
    $orderReference = WC()->session->get("orderReference");

    if (!isset($orderReference)) {
      $orderReference = uniqid('3');
      WC()->session->set("orderReference", $orderReference);
    }

    $coupons = array_map(function($c) {
      $coupon = new WC_Coupon($c);
      return array(
        'amount' => $coupon->get_amount(),
        'type' => $coupon->get_discount_type(),
        'freeShipping' => $coupon->get_free_shipping(),
        'description' => $coupon->get_description(),
      );
    }, $cart->get_applied_coupons());

    $items = array_map(function ($i) {
      $category_ids = $i['data']->get_category_ids();
      $term = sizeof($category_ids) > 0 ? get_term_by('id', $category_ids[0], 'product_cat') : null;
      /* @var $product WC_Product */
      $product = $i['data'];
      $unitPriceWithTaxAndDiscount = (floatval($i['line_total']) + (isset($i['line_tax']) ? floatval($i['line_tax']) : 0)) / $i['quantity'];
      return [
        'name' => $product->get_name(),
        'quantity' => $i['quantity'],
        'unitPrice' => $unitPriceWithTaxAndDiscount,
        'sku' => $product->get_id() . '_' . $i['key'],
        'type' => ($product->get_downloadable() ? 'DIGITAL' : $product->get_virtual()) ? 'SERVICE' : 'PHYSICAL',
        'category' => $term ? $term->name : null,
      ];
    }, array_values($cart->get_cart_contents()));

    if ($cart->needs_shipping())
      $items[] = [
        'name' => $this->paywerk_label('plugin.checkout.shipping', 'Shipping'),
        'sku' => 'shipping',
        'quantity' => 1,
        'unitPrice' => $cart->get_shipping_total() + $cart->get_shipping_tax(),
        'type' => 'DELIVERY'
      ];
    $user = wp_get_current_user();

    $shipToDifferentAddress = isset($customerCheckoutFields['ship_to_different_address']) && $customerCheckoutFields['ship_to_different_address'] == 1;

    $billingAddress = array(
      'countryCode' => $customerCheckoutFields['billing_country'],
      'postalCode' => $customerCheckoutFields['billing_postcode'],
      'region' => $customerCheckoutFields['billing_state'],
      'city' => $customerCheckoutFields['billing_city'],
      'street' => $customerCheckoutFields['billing_address_1'] . ' ' . isset($customerCheckoutFields['billing_address_2']) ? $customerCheckoutFields['billing_address_2'] : '',
    );
    $shippingAddress = $shipToDifferentAddress ? array(
      'countryCode' => $customerCheckoutFields['shipping_country'],
      'postalCode' => $customerCheckoutFields['shipping_postcode'],
      'region' => $customerCheckoutFields['shipping_state'],
      'city' => $customerCheckoutFields['shipping_city'],
      'street' => $customerCheckoutFields['shipping_address_1'] . ' ' . isset($customerCheckoutFields['shipping_address_2']) ? $customerCheckoutFields['shipping_address_2'] : '',
    ) : $billingAddress;

    $params = array(
      'lang' => $this->lang,
      'firstName' => $customerCheckoutFields['billing_first_name'],
      'lastName' => $customerCheckoutFields['billing_last_name'],
      'email' => $customerCheckoutFields['billing_email'],
      'phone' => $customerCheckoutFields['billing_phone'],
      'reference' => $orderReference,
      'billingAddress' => $billingAddress,
      'deliveryAddress' => $shippingAddress,
      'amount' => $cart->total,
      'currency' => get_woocommerce_currency(),
      'successUrl' => $this->get_return_url(),
      'failedUrl' => $this->get_return_url(),
      'callbackUrl' => site_url('?wc-api=paywerk_payment_status'),
      'items' => $items,
      'metadata' =>
        array(
          'other' => [
            'coupons' => $coupons,
            'originalItems' => $cart->get_cart_contents(),
          ],
          'shopPlatform' => 'WooCommerce',
          'platformVersion' => WC_VERSION,
          'pluginVersion' => PAYWERK_VERSION,
        ),
    );

    $isShopperOrderHistoryEnabled = isset($this->settings['sendOrderHistory']) && $this->settings['sendOrderHistory'] == 'yes';

    if ($isShopperOrderHistoryEnabled) {
      $wc_get_customer_total_spent = wc_get_customer_total_spent($user->ID);

      if ($wc_get_customer_total_spent > 0) {
        $lastOrders = $this->loadShopperOrderHistory($customer['id']);
        $params['history'] =
          array(
            'numOrders' => wc_get_customer_order_count($user->ID),
            'totalAmount' => array(array('currency' => get_woocommerce_currency(), 'amount' => $wc_get_customer_total_spent)),
            'lastOrders' => $lastOrders,
          );
      }
    }

    if ($user->user_registered) {
      $params['registeredAt'] = $this->to_iso($user->user_registered) . 'Z';
    }

    $requestUrl = $this->url . '/api/v1/orders';
    $jsonRequest = json_encode($params);

    if (empty($customerCheckoutFields['billing_country'])) return false;

    return wp_remote_post($requestUrl, array(
      'timeout' => 10,
      'headers' => [
        'Content-Type' => 'application/json',
        'Accept' => 'application/json',
        'Accept-Language' => $this->lang,
        'Authorization' => 'Bearer ' . $this->apiKey
      ],
      'body' => $jsonRequest
    ));
  }

  /**
   * @param $id
   * @return array|array[]|null
   */
  public function loadShopperOrderHistory($id) {
    $lastOrders = null;
    if (($id ?? 0)) {
      $lastWCOrders = wc_get_orders(array(
        'customer' => $id,
        'date_created' => '>' . (new DateTime('now'))->modify('-2 years')->format('Y-m-d'),
        'limit' => 100,
        'orderby' => 'date',
      ));

      $completedOrders = array_filter($lastWCOrders, function ($order) {
        return $order->is_paid() || $order->needs_processing();
      });

      try {
        $lastOrders = array_map(function ($o) {
          return [
            'reference' => $o->get_order_number(),
            'amount' => $o->get_total(),
            'currency' => get_woocommerce_currency(),
            'paymentMethod' => $o->get_payment_method(),
            'refunded' => $o->get_total_refunded(),
            'createdAt' => $this->to_iso($o->get_date_created()),
            'status' => $o->get_status(),
            'email' => $o->get_billing_email(),
            'phone' => $o->get_billing_phone(),
            'items' => array_map(function ($i) {
              return [
                'name' => $i->get_name(),
                'quantity' => $i->get_quantity(),
                'unitPrice' => $i->get_total() / $i->get_quantity()
              ];
            }, array_values($o->get_items())),
            'billingAddress' => [
              'countryCode' => $o->get_billing_country(),
              'postalCode' => $o->get_billing_postcode(),
              'region' => $o->get_billing_state(),
              'city' => $o->get_billing_city(),
              'street' => $o->get_billing_address_1()
            ],
            'deliveryAddress' => $o->get_shipping_country() ? [
              'countryCode' => $o->get_shipping_country(),
              'postalCode' => $o->get_shipping_postcode(),
              'region' => $o->get_shipping_state(),
              'city' => $o->get_shipping_city(),
              'street' => $o->get_shipping_address_1()
            ] : null
          ];
        }, $completedOrders);
      } catch (Exception $e) {
        error_log(print_r($e, true));
      }
    }
    return $lastOrders;
  }

  private function isMinAmountInRange() {
    if (!WC()->cart) return false;

    $total = $this->get_order_total();
    return $total > 0 && ($this->min_amount == false || $this->min_amount <= $total);
  }

  public function is_available() {
    return parent::is_available() && $this->isMinAmountInRange();
  }
}
